/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "image.h"

#include <jpeglib.h>
#include <setjmp.h>

/*____________________________________________________________________________*/

struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */
  jmp_buf setjmp_buffer;	/* for return to caller */
};

typedef struct my_error_mgr * my_error_ptr;

/*____________________________________________________________________________*/
static void jpeg_rd_error_exit(j_common_ptr cinfo) {
  /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
  my_error_ptr myerr = (my_error_ptr) cinfo->err;

  /* Always display the message. */
  /* We could postpone this until after returning, if we chose. */
  (*cinfo->err->output_message) (cinfo);

  /* Return control to the setjmp point */
  longjmp(myerr->setjmp_buffer, 1);
}

/*____________________________________________________________________________*/
int read_jpeg(TImage *self) {
  struct jpeg_decompress_struct cinfo;
  struct my_error_mgr jerr;
  JSAMPARRAY buffer;		/* Output row buffer */
  int row_stride;		/* physical row width in output buffer */

  /* Step 1: allocate and initialize JPEG decompression object */
  /* We set up the normal JPEG error routines, then override error_exit. */
	cinfo.err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit = jpeg_rd_error_exit;
  /* Establish the setjmp return context for my_error_exit to use. */
	if (setjmp(jerr.setjmp_buffer)) {
  /* If we get here, the JPEG code has signaled an error.
   * We need to clean up the JPEG object, close the input file, and return. */
Close02:
//		if (decompress_open) jpeg_finish_decompress(&cinfo);
		jpeg_destroy_decompress(&cinfo);
		return False;
	}
  /* Now we can initialize the JPEG decompression object. */
	jpeg_create_decompress(&cinfo);
  /* Step 2: specify data source (eg, a file) */
	jpeg_stdio_src(&cinfo, self->in_fp);

  /* Step 3: read file parameters with jpeg_read_header() */

	jpeg_read_header(&cinfo, TRUE);
  /* We can ignore the return value from jpeg_read_header since
   *   (a) suspension is not possible with the stdio data source, and
   *   (b) we passed TRUE to reject a tables-only JPEG file as an error.
   * See libjpeg.doc for more info.
   */
	/* Set 24-bit RGB output */
	cinfo.out_color_space = JCS_RGB;
	cinfo.quantize_colors = FALSE;

	switch(self->pxm_LD & imgLD_scale_mask) {
	case imgLD_scale_2: cinfo.scale_denom = 2; break;
	case imgLD_scale_4: cinfo.scale_denom = 4; break;
	case imgLD_scale_8: cinfo.scale_denom = 8; break;
	}

	jpeg_calc_output_dimensions(&cinfo);

	self->width = cinfo.output_width;
	self->height = cinfo.output_height;

	if (self->pxm_LD & pxmLD_OnlyInfo) {
	    jpeg_destroy_decompress(&cinfo);
	    return True;
	}
// printf("%s(%d): %dx%d %dx%d  %d %d\n", FL, cinfo.image_width, cinfo.image_height, 
//		cinfo.output_width, cinfo.output_height,
//		cinfo.scale_num, cinfo.scale_denom);


  /* Step 4: set parameters for decompression */

  /* In this example, we don't need to change any of the defaults set by
   * jpeg_read_header(), so we do nothing here.
   */
//	if (!TImage_SetMem(self, gim)) goto Close02;
	if (!TImage_SetMem(self)) goto Close02;

  /* Step 5: Start decompressor */

	jpeg_start_decompress(&cinfo);
  /* JSAMPLEs per row in output buffer */
	row_stride = cinfo.output_width * cinfo.output_components;
  /* Make a one-row-high sample array that will go away when done with image */
	buffer = (*cinfo.mem->alloc_sarray)
		((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

  /* Step 6: while (scan lines remain to be read) */
  /*           jpeg_read_scanlines(...); */

  /* Here we use the library's state variable cinfo.output_scanline as the
   * loop counter, so that we don't have to keep track ourselves.
   */

  bool rgb565 = (self->pxm_LD & pxmLD_bppMask) == pxm_2b_16bpp;
  byte *sb, *db0, *db;
  int x;

	db0 = self->data;
	while (cinfo.output_scanline < cinfo.output_height) {
	    jpeg_read_scanlines(&cinfo, buffer, 1);
	    sb = buffer[0];
	    db = db0;
	    x = cinfo.output_width;
	    if (rgb565) {
		for (;x;x--) {
		    *((word*)db) = ((sb[0] & 0xf8) << 8) |
				   ((sb[1] & 0xfc) << 3) |
				   ((sb[2] & 0xf8) >> 3);
		    db += 2; sb += 3;
		}
	    } else {
		for (;x;x--) {
		    *((dword*)db) = (sb[0] << 16) | (sb[1] << 8) | sb[2];
		    db += 4; sb += 3;
		}
	    }
	    db0 += self->bytes_per_line;
	}


  /* Step 7: Finish decompression */

	jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);

  /* At this point you may want to check to see whether any corrupt-data
   * warnings occurred (test whether jerr.pub.num_warnings is nonzero).
   */
	return True;
}

static bool ChkType(TglobPath *gp, const char *fn, FILE *F) {
	return TImage_ChkPattern(gp, fn, hImageRD_jpeg.pattern);
}

const ThImageLoader hImageRD_jpeg = {
  "jpg jpeg",
  filetyp_JPEG,
  ChkType,
  read_jpeg
};
